import mido

supported_extensions = ['mid', 'midi']


def parse_file(filename):
    """ Returns a list of notes from the note_on events of a MIDI file. """
    mid = mido.MidiFile(filename)
    slices = []
    slic = []
    for msg in mid:
        if msg.type == 'note_on' and msg.velocity > 0:
            if msg.time > 0:
                slices.append(slic)
                slic = [msg.note % 12]
            else:
                slic.append(msg.note % 12)
    if slic:
        slices.append(slic)
    return slices



import justkeydding.dataset
from justkeydding.optimizer import ensembler
import numpy as np
import sys
import joblib
import os


tonic_names = [
    'C', 'C#', 'D', 'Eb',
    'E', 'F', 'F#', 'G',
    'Ab', 'A', 'Bb', 'B',
]


def key_name(key_number):
    mode = 'major' if key_number < 12 else 'minor'
    tonic = tonic_names[key_number % 12]
    return '{}\t{}'.format(tonic, mode)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print('You need to provide an input file')
        exit()
    filename = sys.argv[1]
    model = 'pre-trained.joblib'
    if not os.path.exists(model):
        print('You need to train a model first')
        exit()
    key_profiles = [
        'aarden_essen',
        'temperley',
        'krumhansl_kessler',
        'bellman_budge',
        'albrecht_shanahan1',
        'albrecht_shanahan2',
        'sapp',
        'simple_natural_minor',
        'simple_harmonic_minor',
        'simple_melodic_minor', ]
    key_transitions = [
        'ktg_exponential5',
        'ktg_exponential10',
        'ktg_exponential15', ]
    mixed_profiles = False
    ens = ensembler.Ensembler(key_profiles, key_transitions)
    features = ens.evaluate(filename, mixed_profiles)
    features = [f for l in features for f in l]
    feature_array = np.array(features).reshape(1, -1)
    feature_array = justkeydding.dataset.feature_scaling(feature_array)
    clf = joblib.load(model)
    prediction = clf.predict(feature_array)[0]
    print(key_name(prediction))



from justkeydding.parameters import key_profiles, key_transitions
import random
import logging
import copy


class Evolver:
    def __init__(self, generator):
        self.logger = logging.getLogger('evolver')
        self.generator = generator
        self.logger.info('Evolver() <- generator={}'.format(generator))
        self.kp_id = 0
        self.kt_id = 0

    def evolve_key_profiles(self, kps, retain=0.2, random_retain=0.2, crossover=0.3, mutation_prob=0.5, mutation_ratio=0.1):
        ''' Evolver, assumes key_profiles sorted by best performance first '''
        self.logger.info('Start evolve_key_profiles() <- kps={}, retain={}, random_retain={}, crossover={}, mutation_prob={}, mutation_ratio={}'.format(kps, retain, random_retain, crossover, mutation_prob, mutation_ratio))
        # Number of key profiles from the old population that we are keeping
        retain_length = int(len(kps) * retain)
        self.logger.debug('retain_length:{}'.format(retain_length))
        # Other random key profiles from the population that we are keeping
        random_retain_length = int(len(kps) * random_retain)
        self.logger.debug('random_retain_length:{}'.format(random_retain_length))
        # Number of key profiles that will be generated by mixing
        # the major and minor key profiles of 2 different samples
        crossover_length = int(len(kps) * crossover)
        self.logger.debug('crossover_length:{}'.format(crossover_length))

        # Initialize the parents of the new population
        new_key_profiles = copy.deepcopy(kps[:retain_length])

        # Select random key profiles from the previous generation
        random_old = []
        while len(random_old) < random_retain_length:
            kp = random.choice(kps[retain_length:])
            if kp not in random_old and kp not in new_key_profiles:
                self.logger.debug('random_kp_retained:{}'.format(kp))
                random_old.append(copy.deepcopy(kp))
        new_key_profiles.extend(random_old)

        # Crossover parents to fill some of the population
        children = []
        while len(children) < crossover_length:
            major_parent = random.choice(new_key_profiles)
            minor_parent = random.choice(new_key_profiles)
            if major_parent != minor_parent:
                # Child is made of the major key of
                # the male and minor key of the female
                child_name = 'kpc{}'.format(self.kp_id)
                self.kp_id += 1
                major_kp = key_profiles.get(major_parent)
                minor_kp = key_profiles.get(minor_parent)
                child_kp = major_kp[:12] + minor_kp[12:]
                self.logger.debug('child:{}, major_parent:{}, minor_parent:{}'.format(child_name, major_parent, minor_parent))
                key_profiles.insert_new(child_name, child_kp)
                children.append(child_name)
        new_key_profiles.extend(children)

        # Do some mutations. Mutations consist in taking a 'fraction'
        # of the magnitude of a pitch class and give it to another pitch class
        # The fraction is defined by 'mutation_ratio', default is 10%
        for idx, key_profile_name in enumerate(new_key_profiles):
            if mutation_prob > random.random():
                mutation_name = 'kpm{}'.format(self.kp_id)
                self.logger.debug('mutation:{}, parent: {}'.format(mutation_name, key_profile_name))
                self.kp_id += 1
                kp = copy.deepcopy(key_profiles.get(key_profile_name))
                index_range = random.choice([(0, 11), (12, 23)])
                pc1_index = 0
                pc2_index = 0
                while pc1_index == pc2_index:
                    pc1_index = random.randint(index_range[0], index_range[1])
                    pc2_index = random.randint(index_range[0], index_range[1])
                delta = kp[pc1_index] * mutation_ratio
                self.logger.debug('pc1 {} changes by {}, from {} to {}'.format(pc1_index, delta, kp[pc1_index], kp[pc1_index] - delta))
                self.logger.debug('pc2 {} changes by {}, from {} to {}'.format(pc2_index, delta, kp[pc2_index], kp[pc2_index] + delta))
                kp[pc1_index] -= delta
                kp[pc2_index] += delta
                key_profiles.insert_new(mutation_name, kp)
                new_key_profiles[idx] = mutation_name

        # Fill the rest of the population with random key profiles
        while len(kps) > len(new_key_profiles):
            kp_name = self.generator.generate_key_profile()
            new_key_profiles.append(kp_name)
        self.logger.info('Done evolve_key_profiles() -> new_key_profiles={}'.format(new_key_profiles))
        return new_key_profiles

    def evolve_key_transitions(self, kts, retain=0.2, random_retain=0.2, mutation_prob=1, mutation_ratio=0.1):
        ''' Evolver, assumes key_transitions sorted by best performance first'''
        self.logger.info('Start evolve_key_transitions() <- kts={}, retain={}, random_retain={}, mutation_prob={}, mutation_ratio={}'.format(kts, retain, random_retain, mutation_prob, mutation_ratio))

        # Number of key transitions from the old population that we are keeping
        retain_length = int(len(kts) * retain)
        self.logger.debug('retain_length:{}'.format(retain_length))
        # Other random key transitions from the population that we are keeping
        random_retain_length = int(len(kts) * random_retain)
        self.logger.debug('random_retain_length:{}'.format(random_retain_length))

        # Initialize the parents of the new population
        new_key_transitions = copy.deepcopy(kts[:retain_length])

        # Select random key transitions from the previous generation
        random_old = []
        while len(random_old) < random_retain_length:
            kt = random.choice(kts[retain_length:])
            if kt not in random_old and kt not in new_key_transitions:
                self.logger.debug('random_kt_retained:{}'.format(kt))
                random_old.append(copy.deepcopy(kt))
        new_key_transitions.extend(random_old)

        # Do some mutations.
        for idx, key_transition_name in enumerate(new_key_transitions):
            if key_transition_name.startswith('ktg'):
                if mutation_prob > random.random():
                    mutation_name = 'ktgm{}'.format(self.kt_id)
                    self.kt_id += 1
                    ratio = key_transitions.get_ratio(key_transition_name)
                    new_ratio = random.uniform(ratio - ratio*mutation_ratio, ratio + ratio*mutation_ratio)
                    kt = self.generator.generate_geometric_key_transition(new_ratio)
                    self.logger.debug('mutation:{}, parent:{}, original_ratio={}, mutation_ratio={}'.format(mutation_name, key_transition_name, ratio, new_ratio))
                    key_transitions.store_ratio(mutation_name, new_ratio)
                    key_transitions.insert_new(mutation_name, kt)
                    new_key_transitions[idx] = mutation_name
            elif key_transition_name.startswith('kts'):
                if mutation_prob > random.random():
                    mutation_name = 'ktsm{}'.format(self.kt_id)
                    self.kt_id += 1
                    self.logger.debug('mutation:{}, parent:{}'.format(mutation_name, key_transition_name))
                    kt = copy.deepcopy(key_transitions.get(key_transition_name))
                    key1_index = 0
                    key2_index = 0
                    while key1_index == key2_index:
                        key1_index = random.randint(0, 23)
                        key2_index = random.randint(0, 23)
                    delta = kt[key1_index] * mutation_ratio
                    self.logger.debug('key1 {} changes by {}, from {} to {}'.format(key1_index, delta, kt[key1_index], kt[key1_index] - delta))
                    self.logger.debug('key2 {} changes by {}, from {} to {}'.format(key2_index, delta, kt[key2_index], kt[key2_index] + delta))
                    kt[key1_index] -= delta
                    kt[key2_index] += delta
                    key_transitions.insert_new(mutation_name, kt)
                    new_key_transitions[idx] = mutation_name

        # Fill the rest of the population with random key transitions
        while len(kts) > len(new_key_transitions):
            kt_name = self.generator.generate_key_transition()
            new_key_transitions.append(kt_name)
        self.logger.info('Done evolve_key_transitions() -> new_key_transitions={}'.format(new_key_transitions))
        return new_key_transitions




"""justkeydding

Key-finding algorithm for Symbolic and Audio files.

Nestor Napoles (napulen@gmail.com)
"""

import justkeydding.inputs
import justkeydding.parameters
import justkeydding.parameters.key_transitions as kt
import justkeydding.parameters.key_profiles as kp
from justkeydding.parameters.observations import Observations
import justkeydding.cli

import numpy as np
from collections import Counter


def create_transition_probabilities(key_transitions):
    """Returns the transition probabilities"""
    d = dict()
    for idx, key in enumerate(justkeydding.parameters.states):
        tonic = key_transitions[:12]
        relative = key_transitions[12:]
        tonic_rotation = -(idx % 12)
        relative_rotation = -(idx % 12)
        if idx >= 12:
            tonic, relative = relative, tonic
            tonic_rotation = (tonic_rotation - 3) % 12
        probs1 = tonic[tonic_rotation:] + tonic[:tonic_rotation]
        probs2 = relative[relative_rotation:] + relative[:relative_rotation]
        kt_ = probs1 + probs2
        d[key] = {key: kt_[idx] for idx, key in enumerate(justkeydding.parameters.states)}
    return d


def create_emission_probabilities(major, minor):
    """Returns the emission probabilities"""
    d = dict()
    for idx, key in enumerate(justkeydding.parameters.states):
        rotation = -(idx % 12)
        profile = major if idx < 12 else minor
        profile = profile[rotation:] + profile[:rotation]
        d[key] = {pc: profile[pc] for pc in range(12)}
    return d


def resolve_input_type(input_file):
    extension = input_file.rsplit('.')[-1]
    input_type = ''
    if extension in justkeydding.inputs.midi.supported_extensions:
        input_type = 'midi'
    elif extension in justkeydding.inputs.symbolic.supported_extensions:
        input_type = 'symbolic'
    elif extension in justkeydding.inputs.audio.supported_extensions:
        input_type = 'audio'
    return input_type


def extract_input_sequence(input_file, is_sequence=False):
    if is_sequence:
        # TODO: Implement this for new format of input sequences
        input_sequence = [int(s) for s in input_file.split(',')]
    else:
        input_sequence = input_format.parse_file(input_file)
    return Observations(input_sequence)


def mylog(x):
    """Returns the logarithm of x (without the annoying warnings of np.log)"""
    return np.log(x) if x > 8.7565e-27 else -np.inf


def viterbi(obs, states, start_p, trans_p, emit_p):
    V = [{}]
    for st in states:
        V[0][st] = {
            "prob": mylog(start_p[st])
            + mylog(emit_p[st][obs[0]]), "prev": None
            }
    # Run Viterbi when t > 0
    for t in range(1, len(obs)):
        V.append({})
        for st in states:
            max_tr_prob = max(V[t-1][prev_st]["prob"] + mylog(trans_p[prev_st][st]) for prev_st in states)
            for prev_st in states:
                if V[t-1][prev_st]["prob"] + mylog(trans_p[prev_st][st]) == max_tr_prob:
                    max_prob = max_tr_prob + mylog(emit_p[st][obs[t]])
                    V[t][st] = {"prob": max_prob, "prev": prev_st}
                    break
    opt = []
    # The highest probability
    max_prob = max(value["prob"] for value in V[-1].values())
    previous = None
    # Get most probable state and its backtrack
    for st, data in V[-1].items():
        if data["prob"] == max_prob:
            opt.append(st)
            previous = st
            break
    # Follow the backtrack till the first observation
    for t in range(len(V) - 2, -1, -1):
        opt.insert(0, V[t + 1][previous]["prev"])
        previous = V[t + 1][previous]["prev"]
    return opt, max_prob


def analyze(input_sequence, kp_major_name, kp_minor_name, kt_name):
    # Preparing the args for the first HMM
    key_transition = kt._kt[kt_name]
    trans_p = create_transition_probabilities(key_transition)
    major = kp._kp[kp_major_name][:12]
    minor = kp._kp[kp_minor_name][12:]
    emit_p = create_emission_probabilities(major, minor)
    obs = input_sequence
    local_keys, max_p = viterbi(
        obs,
        justkeydding.parameters.states,
        justkeydding.parameters.start_p,
        trans_p,
        emit_p)
    sliced_local_keys = list(zip(local_keys, obs.slice_indexes()))
    sliced_local_keys = Observations.from_tuples(sliced_local_keys)
    # Preparing the args for the second HMM
    obs = local_keys  # the keys become the observations
    emit_p = trans_p  # the transitions become emission
    key_transitions = kt._kt["null"]
    trans_p = create_transition_probabilities(key_transitions)
    key, max_prob = viterbi(
        obs,
        justkeydding.parameters.states,
        justkeydding.parameters.start_p,
        trans_p,
        emit_p
    )
    global_key = key[0]
    return [global_key, sliced_local_keys]


def postprocess_local_keys(local_keys):
    slices = []
    for slic in local_keys.slices:
        if slic:
            c = Counter(slic)
            slices.append(max(slic, key=c.get))
        else:
            slices.append('')
    return slices


if __name__ == '__main__':
    args = justkeydding.cli.parse_args()
    input_type = resolve_input_type(args.input)
    if input_type == 'midi':
        import justkeydding.inputs.midi as input_format
    elif input_type == 'symbolic':
        import justkeydding.inputs.symbolic as input_format
    elif input_type == 'audio':
        import justkeydding.inputs.audio as input_format
    elif not args.is_sequence:
        raise ValueError("unsuported file type")
    input_sequence = extract_input_sequence(
        args.input,
        args.is_sequence
    )
    outputs = analyze(
        input_sequence,
        args.key_profile_major,
        args.key_profile_minor,
        args.key_transition
    )
    if args.output_local:
        keys_by_onset = postprocess_local_keys(outputs[1])
        print('{}\n{}'.format(outputs[0], keys_by_onset))
        if args.annotate_file:
            input_format.annotate_local_keys(args.input, keys_by_onset)
    else:
        print(outputs[0])



import argparse


def parse_args():
    parser = argparse.ArgumentParser(
        description='justkeydding for symbolic music files, python version')
    parser.add_argument(
        'input',
        help='Input symbolic music file (or folder if --batch)'
    )
    parser.add_argument(
        '--sequence',
        dest='is_sequence',
        const=True,
        action='store_const',
        default=False,
        help='Provide the input as a string of comma-separated pitch-classes'
    )
    parser.add_argument(
        '--local',
        dest='output_local',
        const=True,
        action='store_const',
        help='Output local keys'
    )
    parser.add_argument(
        '--transition',
        dest='key_transition',
        choices=[
            'ktg_exponential10',
            'ktg_exponential2'
        ],
        default='ktg_exponential10',
        help='Key transition to use'
    )
    parser.add_argument(
        '--majorEmission',
        dest='key_profile_major',
        choices=[
            'krumhansl_kessler',
            'aarden_essen',
            'sapp',
            'bellman_budge',
            'temperley',
        ],
        default='sapp',
        help='Major key profile to use as emission probability distribution'
    )
    parser.add_argument(
        '--minorEmission',
        dest='key_profile_minor',
        choices=[
            'krumhansl_kessler',
            'aarden_essen',
            'sapp',
            'bellman_budge',
            'temperley',
        ],
        default='sapp',
        help='Minor key profile to use as emission probability distribution'
    )
    parser.add_argument(
        '--annotate',
        dest='annotate_file',
        const=True,
        action='store_const',
        help='Annotate the input file'
    )
    args = parser.parse_args()
    return args



import numpy as np
from sklearn.preprocessing import MinMaxScaler
import collections
import re
import os


class Dataset():
    def __init__(self, dataset_name):
        self.dataset_dir = 'datasets'
        self.dataset_name = dataset_name
        if not os.path.exists(self.dataset_dir):
            print('Could not find the dataset dir')
            exit()
        files = []
        for f in os.listdir(self.dataset_dir):
            if f.startswith(self.dataset_name):
                files.append(f)
        annotations_re = re.escape(self.dataset_name) + '-?(.*)_annotations.pkl'
        features_re = re.escape(self.dataset_name) + '-?(.*)_features.pkl'
        ensemble_file = '{}_ensemble.txt'.format(self.dataset_name)
        for f in files:
            # print('Loading {}...'.format(f))
            if f == ensemble_file:
                with open(os.path.join(self.dataset_dir, f)) as fd:
                    lines = fd.readlines()
                if len(lines) != 3:
                    print('The format of the ensemble metadata is incorrect.')
                    exit()
                kps, kts, mixed_profiles = lines
                kps = kps.split(',')
                kps = [s.strip() for s in kps]
                kts = kts.split(',')
                kts = [s.strip() for s in kts]
                mixed_profiles = mixed_profiles.split('=')[1] == "True"
                ensemble = {
                    'key_profiles': kps,
                    'key_transitions': kts,
                    'mixed_profiles': mixed_profiles
                }
                setattr(self, 'ensemble', ensemble)
            match = re.search(features_re, f, re.IGNORECASE)
            if match:
                split_name = match.group(1)
                array = np.load(
                    os.path.join(self.dataset_dir, f),
                    allow_pickle=True
                )
                if split_name:
                    attr_name = '{}_X'.format(split_name)
                    setattr(self, attr_name, array)
                else:
                    self.X = array
            else:
                match = re.search(annotations_re, f, re.IGNORECASE)
                if match:
                    split_name = match.group(1)
                    array = np.load(
                        os.path.join(self.dataset_dir, f),
                        allow_pickle=True
                    )
                    if split_name:
                        attr_name = '{}_y'.format(split_name)
                        setattr(self, attr_name, array)
                    else:
                        self.y = array


def transpose(Xi, y):
    is_minor = y // 12
    trans_features = []
    trans_labels = []
    for ti in range(12):
        Xip = Xi.reshape(-1, 12)
        label = ((y + ti) % 12) + (12 * is_minor)
        features = np.roll(Xip, ti, axis=1)
        trans_features.append(features)
        trans_labels.append(label)
    trans_features = np.array(trans_features).reshape(12, -1)
    trans_labels = np.array(trans_labels)
    return np.array(trans_features), np.array(trans_labels)


def compare_keys(k1, k2):
    if k2 == k1:
        return 'tonic'
    if k1 < 12:
        dominant = (k1 + 7) % 12
        subdominant = (k1 + 5) % 12
        relative = ((k1 + 9) % 12) + 12
        parallel = k1 + 12
    else:
        dominant = ((k1 + 7) % 12) + 12
        subdominant = ((k1 + 5) % 12) + 12
        relative = (k1 - 9) % 12
        parallel = k1 - 12
    if k2 == dominant:
        return 'fifth'
    elif k2 == subdominant:
        return 'fifth'
    elif k2 == relative:
        return 'relative'
    elif k2 == parallel:
        return 'parallel'
    else:
        return 'other'


def weighted_score(pred, y):
    length = len(pred)
    if length != len(y):
        print('These arrays should have the same number of elements')
        return
    keys = [compare_keys(y[i], pred[i]) for i in range(length)]
    cnt = collections.Counter(keys)
    tonics = cnt.get('tonic', 0)
    fifths = cnt.get('fifth', 0)
    relatives = cnt.get('relative', 0)
    parallels = cnt.get('parallel', 0)
    others = cnt.get('other', 0)
    mirex_score = tonics + 0.5*fifths + 0.3*relatives + 0.2*parallels
    all_scores = [mirex_score, tonics, fifths, relatives, parallels, others]
    all_scores = [x / length for x in all_scores]
    return all_scores


def data_augmentation(X, y):
    original_length = X.shape[1]
    trans_features = []
    trans_labels = []
    for i in range(len(X)):
        Xp, yp = transpose(X[i], y[i])
        trans_features.append(Xp)
        trans_labels.append(yp)
    trans_features = np.array(trans_features).reshape(-1, original_length)
    trans_labels = np.array(trans_labels).reshape(-1)
    return trans_features, trans_labels


def feature_scaling(X, minmax=(-1, 1)):
    scaler = MinMaxScaler(feature_range=minmax)
    original_length = X.shape[1]
    Xp = X.reshape(-1, 24)
    Xp = np.apply_along_axis(
        lambda x: scaler.fit_transform(x.reshape(-1, 1)), 1, Xp
    )
    Xp = Xp.reshape(-1, original_length)
    return Xp



import justkeydding.dataset
from sklearn.linear_model import LogisticRegression
import numpy as np
import joblib

if __name__ == '__main__':

    clf_name = 'pre-trained'

    dataset_filenames = [
        'a-maps_chroma',
        # 'a-maps_symbolic',
        'albrecht_chroma',
        # 'albrecht_symbolic',
        'arthur2018-recordings_chroma',
        'arthur2018-synth-shortname_chroma',
        # 'arthur2018-synth-shortname_symbolic',
        'billboard_chroma',
        'cross-era-shortname_chroma',
        'giantsteps-key_chroma',
        'giantsteps-mtg-key_chroma',
        'napoles2018_chroma',
        # 'napoles2018_symbolic',
    ]

    for idx, d in enumerate(dataset_filenames):
        data = justkeydding.dataset.Dataset(d)
        if idx == 0:
            ensemble = data.ensemble
            fulldataset_X = data.X
            fulldataset_y = data.y
        else:
            if data.ensemble != ensemble:
                print('This dataset is incompatible with other datasets')
                exit()
            fulldataset_X = np.concatenate((fulldataset_X, data.X))
            fulldataset_y = np.concatenate((fulldataset_y, data.y))

    clf = LogisticRegression(
        penalty='l2',
        fit_intercept=False,
        dual=False,
        C=0.7,
        solver='lbfgs',
        multi_class='auto',
        max_iter=100000
    )

    # Feature scaling in all the feature vectors
    fulldataset_X = justkeydding.dataset.feature_scaling(fulldataset_X)

    # Data augmentation on the training sets
    fulldataset_X, fulldataset_y = justkeydding.dataset.data_augmentation(
        fulldataset_X,
        fulldataset_y)

    # Training
    clf.fit(fulldataset_X, fulldataset_y)

    # Save the model
    clf_filename = '{}.joblib'.format(clf_name)
    joblib.dump(clf, clf_filename)
